from typing import List, TYPE_CHECKING, Union

from zuper_typing import dataclass

from aido_schemas import FriendlyPose
from .basics import InteractionProtocol

__all__ = [
    "protocol_collision_checking",
    "Circle",
    "Rectangle",
    "Primitive",
    "MapDefinition",
    "CollisionCheckQuery",
    "CollisionCheckResult",
    "PlacedPrimitive",
    "Point",
]

if TYPE_CHECKING:
    from dataclasses import dataclass


@dataclass
class Circle:
    radius: float


@dataclass
class Point:
    x: float
    y: float


@dataclass
class Rectangle:
    xmin: float
    ymin: float
    xmax: float
    ymax: float


Primitive = Union[Circle, Rectangle]


@dataclass
class PlacedPrimitive:
    pose: FriendlyPose
    primitive: Primitive


@dataclass
class MapDefinition:
    environment: List[PlacedPrimitive]
    body: List[PlacedPrimitive]


@dataclass
class CollisionCheckQuery:
    pose: FriendlyPose


@dataclass
class CollisionCheckResult:
    collision: bool


protocol_collision_checking = InteractionProtocol(
    description="""Collision checking protocol""",
    inputs={"set_params": MapDefinition, "query": CollisionCheckQuery},
    outputs={"response": CollisionCheckResult},
    language="""
        (in:set_params ; (in:query ; out:response)*)*
        """,
)



from typing import List

from geometry import SE2, SE2value
from zuper_commons.logs import ZLogger

from duckietown_world import pose_from_friendly
from duckietown_world.utils import SE2_apply_R2
from .collision_protocol import Circle, PlacedPrimitive, Rectangle

logger = ZLogger(__name__)

__all__ = ["plot_geometry"]


# noinspection PyListCreation
def plot_geometry(ax, se0: SE2value, structure: List[PlacedPrimitive], style, zorder: int):
    from matplotlib import pyplot

    ax: pyplot.Axes
    for pp in structure:
        if isinstance(pp.primitive, Circle):
            x, y = pp.pose.x, pp.pose.y
            x, y = SE2_apply_R2(se0, (x, y))
            ax.add_artist(pyplot.Circle((x, y), pp.primitive.radius, fill=True, color=style, zorder=zorder))
        if isinstance(pp.primitive, Rectangle):
            q = SE2.multiply(se0, pose_from_friendly(pp.pose))
            points = []

            points.append(SE2_apply_R2(q, (pp.primitive.xmin, pp.primitive.ymin)))
            points.append(SE2_apply_R2(q, (pp.primitive.xmin, pp.primitive.ymax)))
            points.append(SE2_apply_R2(q, (pp.primitive.xmax, pp.primitive.ymax)))
            points.append(SE2_apply_R2(q, (pp.primitive.xmax, pp.primitive.ymin)))
            points.append(SE2_apply_R2(q, (pp.primitive.xmin, pp.primitive.ymin)))
            xs = [_[0] for _ in points]
            ys = [_[1] for _ in points]

            pyplot.fill(xs, ys, "-", color=style, zorder=zorder)
        # if isinstance(pp.primitive, Rectanlge):
        #     x, y = pp.pose.x, pp.pose.y
        #     ax.add_artist(pyplot.Circle((x, y), pp.primitive.radius, fill=False, color=style))



# noinspection PyUnresolvedReferences
from zuper_nodes import InteractionProtocol, particularize, particularize_no_check

# noinspection PyUnresolvedReferences
from zuper_nodes_wrapper import wrap_direct, Context



__version__ = "6.2.18"

from zuper_commons.logs import ZLogger

logger = ZLogger(__name__)
import os

path = os.path.dirname(os.path.dirname(__file__))
logger.debug(f"dt-protocols version {__version__} path {path}")

from .basics import *
from .collision_drawing import *
from .collision_protocol import *



from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Mapping, Type, TypeVar

import yaml
from zuper_commons.fs import locate_files, read_ustring_from_utf8_file
from zuper_ipce import object_from_ipce

import duckietown_challenges as dc
from zuper_nodes import (
    ExternalProtocolViolation,
    IncompatibleProtocol,
    InteractionProtocol,
)
from zuper_nodes_wrapper.wrapper_outside import ComponentInterface
from . import logger

__all__ = ["run_checker"]

Y = TypeVar("Y")
S = TypeVar("S")


@dataclass
class CheckerSession:
    dataset: object
    scores: List
    responses: List

def run_checker(
    cie: dc.ChallengeInterfaceEvaluator,
    protocol: InteractionProtocol,
    *,
    dirname: str,
    K: Type[Y],
    scoring: Callable[[Y, Any], S],
    finalize_scores: Callable[[List[S]], Mapping[str, float]],
) -> Dict[str, CheckerSession]:
    agent_ci = ComponentInterface(
        fnin="/fifos/checker-in",
        fnout="/fifos/checker-out",
        expect_protocol=protocol,
        nickname="checker",
    )
    try:

        # check compatibility so that everything
        # fails gracefully in case of error
        # noinspection PyProtectedMember
        try:
            agent_ci._get_node_protocol()
        except IncompatibleProtocol as e:
            msg = "Invalid protocol"
            raise dc.InvalidSubmission(msg) from e

        K_params = protocol.inputs["set_params"]
        K_query = protocol.inputs["query"]

        # K_response = protocol.outputs["response"]

        @dataclass
        class Interaction:
            query: K_query
            gt: K

        @dataclass
        class Data:
            params: K_params
            interactions: List[Interaction]

        a = locate_files(dirname, "*.tests.yaml")
        scores = []
        responses = []
        episodes = {}
        for fn in a:
            data = read_ustring_from_utf8_file(fn)
            ydata = yaml.load(data, Loader=yaml.Loader)
            inside = object_from_ipce(ydata, Data)
            logger.info(fn=fn)
            agent_ci.write_topic_and_expect_zero("set_params", inside.params)
            for interaction in inside.interactions:
                q = interaction.query
                r = interaction.gt
                msg = agent_ci.write_topic_and_expect("query", q, expect="response")
                response = msg.data
                scores.append(scoring( r, response))
                responses.append(response)

            episodes[fn] = CheckerSession(dataset=inside, scores=scores, responses=responses)
        final_scores = finalize_scores(scores)

        for k, v in final_scores.items():
            cie.set_score(k, v)

    except ExternalProtocolViolation as e:
        msg = "The remote node has violated protocol"
        raise dc.InvalidSubmission(msg) from e
    except dc.InvalidSubmission:
        raise
    except BaseException as e:
        raise dc.InvalidEvaluator() from e

    finally:
        agent_ci.close()

    return episodes




from setuptools import setup


def get_version(filename: str):
    import ast

    version = None
    with open(filename) as f:
        for line in f:
            if line.startswith("__version__"):
                version = ast.parse(line).body[0].value.s
                break
        else:
            raise ValueError("No version found in %r." % filename)
    if version is None:
        raise ValueError(filename)
    return version


version = get_version(filename="src/dt_protocols/__init__.py")

line = "daffy"
install_requires = [
    "aido-protocols-daffy",
    "termcolor",
    "zuper-nodes-z6>=6.0.37",
    "PyGeometry-z6",
    "duckietown-world-daffy",
]

setup(
    name=f"dt-protocols-{line}",
    version=version,
    keywords="",
    package_dir={"": "src"},
    packages=["dt_protocols"],
    install_requires=install_requires,
    entry_points={
        "console_scripts": [],
    },
)
